@@ -327,24 +327,45 @@ activation_clients_cache_notify (void)
 gboolean
 activation_clients_is_empty_scan (void)
 {
-        GList *l, *next;
+        GList *l, *check = NULL;
 
-        for (l = clients; l; l = next) {
-                next = l->next;
+        for (l = clients; l; l = l->next)
+                check = g_list_prepend (check, CORBA_Object_duplicate (l->data, NULL));
+
+        for (l = check; l; l = l->next) {
                 if (ORBit_small_get_connection_status (l->data) ==
                     ORBIT_CONNECTION_DISCONNECTED) {
-                        CORBA_Object_release (l->data, NULL);
-                        clients = g_list_delete_link (clients, l);
+                        GList *remove;
+                        if ((remove = g_list_find (clients, l->data))) {
+                                CORBA_Object_release (remove->data, NULL);
+                                clients = g_list_delete_link (clients, remove);
+                        }
                 }
+                CORBA_Object_release (l->data, NULL);
         }
 
+        g_list_free (check);
+
         return clients == NULL;
 }
 
-static void
-active_client_cnx_broken (ORBitConnection *cnx,
-                          gpointer         dummy)
+static gboolean
+ac_rescan (gpointer data)
 {
+        static gboolean in_rescan = FALSE;
+        static guint idle_id = 0;
+
+        server_lock ();
+
+        /* We tend to get a lot of 'broken' callbacks at once */
+        if (in_rescan) {
+                if (!idle_id)
+                        idle_id = g_idle_add (ac_rescan, NULL);
+                server_unlock ();
+                return FALSE;
+        }
+        in_rescan = TRUE;
+
         if (activation_clients_is_empty_scan ()) {
 #ifdef BONOBO_ACTIVATION_DEBUG
                 g_warning ("All clients dead");
@@ -352,6 +373,17 @@ active_client_cnx_broken (ORBitConnection *cnx,
                 check_quit ();
         }
 
+        in_rescan = FALSE;
+        server_unlock ();
+
+        return FALSE;
+}
+
+static void
+active_client_cnx_broken (ORBitConnection *cnx,
+                          gpointer         dummy)
+{
+        ac_rescan (NULL);
 }
 
 static void
@@ -364,6 +396,8 @@ impl_Bonobo_ActivationContext_addClient (PortableServer_Servant        servant,
         gboolean new_locale;
         ORBitConnection *cnx;
 
+        server_lock ();
+
         new_locale = register_interest_in_locales (locales);
 
         cnx = ORBit_small_get_connection (client);
@@ -375,24 +409,28 @@ impl_Bonobo_ActivationContext_addClient (PortableServer_Servant        servant,
                 clients, CORBA_Object_duplicate (client, ev));
 
         if (!l) {
-                g_signal_connect (
-                        cnx, "broken",
-                        G_CALLBACK (active_client_cnx_broken),
-                        NULL);
+                ORBit_small_listen_for_broken
+                        (client, G_CALLBACK (active_client_cnx_broken), NULL);
                 check_quit ();
         }
 
         if (new_locale)
                 bonobo_object_directory_reload ();
+
+        server_unlock ();
 }
 
 static Bonobo_ObjectDirectoryList *
 impl_Bonobo_ActivationContext__get_directories (PortableServer_Servant servant,
                                                 CORBA_Environment     *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
 	Bonobo_ObjectDirectoryList *retval;
 
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
+
 	retval = Bonobo_ObjectDirectoryList__alloc ();
         if (actx->obj != CORBA_OBJECT_NIL) {
                 retval->_length = 1;
@@ -405,6 +443,8 @@ impl_Bonobo_ActivationContext__get_directories (PortableServer_Servant servant,
 
 	CORBA_sequence_set_release (retval, CORBA_TRUE);
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -413,7 +453,11 @@ impl_Bonobo_ActivationContext_addDirectory (PortableServer_Servant servant,
                                             Bonobo_ObjectDirectory dir,
                                             CORBA_Environment     *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
+
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
 
         if (actx->obj == dir)
                 CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
@@ -423,6 +467,8 @@ impl_Bonobo_ActivationContext_addDirectory (PortableServer_Servant servant,
                 CORBA_Object_release (actx->obj, ev);
                 actx->obj = CORBA_Object_duplicate (dir, ev);
         }
+
+        server_unlock ();
 }
 
 static void
@@ -430,7 +476,11 @@ impl_Bonobo_ActivationContext_removeDirectory (PortableServer_Servant servant,
                                                Bonobo_ObjectDirectory dir,
                                                CORBA_Environment     *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
+
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
 
         if (dir != actx->obj)
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
@@ -443,6 +493,8 @@ impl_Bonobo_ActivationContext_removeDirectory (PortableServer_Servant servant,
                 } else
                         directory_info_free (actx, ev);
         }
+
+        server_unlock ();
 }
 
 static void
@@ -458,6 +510,7 @@ ac_do_activation (ActivationContext                  *actx,
 {
 	int num_layers;
 	Bonobo_ServerInfo *activatable;
+        gchar *aid = NULL;
 
 	/* When doing checks for shlib loadability, we 
          * have to find the info on the factory object in case
@@ -500,12 +553,15 @@ ac_do_activation (ActivationContext                  *actx,
 	 * order for loading to work properly (no, we're not going to
 	 * bother with loading a remote shlib into a process - it gets far too complicated
 	 * far too quickly :-) */
+
+        if (activatable)
+                aid = g_strdup_printf ("OAFAID:[%s,%s,%s]", activatable->iid,
+                                       activatable->username, activatable->hostname);
 	
 	if (activatable && !strcmp (activatable->server_type, "shlib")
 	    && !(flags & Bonobo_ACTIVATION_FLAG_NO_LOCAL)
 	    && (hostname && !strcmp (activatable->hostname, hostname))) {
 		int j;
-		char tbuf[512];
 		
 		out->res._d = Bonobo_ACTIVATION_RESULT_SHLIB;		
 
@@ -531,35 +587,29 @@ ac_do_activation (ActivationContext                  *actx,
 
 		/* Copy location into last buffer slot for use in later activation */
 		out->res._u.res_shlib._buffer[j+1] = CORBA_string_dup (activatable->location_info);
-		
-		g_snprintf (tbuf, sizeof (tbuf), "OAFAID:[%s,%s,%s]",
-			    activatable->iid,
-			    activatable->username,
-			    activatable->hostname);
-		out->aid = CORBA_string_dup (tbuf);
+
+		out->aid = CORBA_string_dup (aid);
 	} else {
 		CORBA_Object retval;
+                char *iid = g_strdup (server->iid);
 
 		retval = Bonobo_ObjectDirectory_activate (
-                        actx->obj, server->iid, BONOBO_OBJREF (actx),
+                        actx->obj, iid, BONOBO_OBJREF (actx),
                         environment, flags, client, ctx, ev);
+                g_free (iid);
 
 		if (ev->_major == CORBA_NO_EXCEPTION) {
-			char tbuf[512];
 			out->res._d = Bonobo_ACTIVATION_RESULT_OBJECT;
 			out->res._u.res_object = retval;
-			g_snprintf (tbuf, sizeof (tbuf),
-				    "OAFAID:[%s,%s,%s]", activatable->iid,
-				    activatable->username,
-				    activatable->hostname);
-			out->aid = CORBA_string_dup (tbuf);
+			out->aid = CORBA_string_dup (aid);
 		}
 #ifdef BONOBO_ACTIVATION_DEBUG
                 else
                         g_warning ("Activation of '%s' failed with exception '%s'",
-                                   activatable->iid, ev->_id);
+                                   aid, ev->_id);
 #endif
 	}
+        g_free (aid);
 }
 
 
@@ -574,12 +624,16 @@ impl_Bonobo_ActivationContext_activateMatchingFull (
         CORBA_Context                       ctx,
         CORBA_Environment                  *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
 	Bonobo_ActivationResult *retval = NULL;
 	Bonobo_ServerInfo **items, *curitem;
 	int i;
 	char *hostname;
 
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
+
 	ac_update_lists (actx, ev);
 
 	actx->refs++;
@@ -596,6 +650,7 @@ impl_Bonobo_ActivationContext_activateMatchingFull (
 	retval = Bonobo_ActivationResult__alloc ();
 	retval->res._d = Bonobo_ACTIVATION_RESULT_NONE;
 
+        bonobo_object_ref (actx);
 	for (i = 0; (retval->res._d == Bonobo_ACTIVATION_RESULT_NONE) && items[i]
 	     && (i < actx->total_servers); i++) {
 		curitem = items[i];
@@ -603,6 +658,7 @@ impl_Bonobo_ActivationContext_activateMatchingFull (
 		ac_do_activation (actx, curitem, environment,
 				  retval, flags, hostname, client, ctx, ev);
 	}
+        bonobo_object_unref (actx);
 
 	if (retval->res._d == Bonobo_ACTIVATION_RESULT_NONE)
 		retval->aid = CORBA_string_dup ("");
@@ -617,6 +673,8 @@ impl_Bonobo_ActivationContext_activateMatchingFull (
 
 	actx->refs--;
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -641,12 +699,16 @@ impl_Bonobo_ActivationContext_query (PortableServer_Servant servant,
                                      const Bonobo_StringList * selection_order,
                                      CORBA_Context ctx, CORBA_Environment * ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
+        ActivationContext *actx;
 	Bonobo_ServerInfoList *retval;
 	Bonobo_ServerInfo **items;
 	int item_count;
 	int i, j, total;
 
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
+
 	retval = Bonobo_ServerInfoList__alloc ();
 	retval->_length = 0;
 	retval->_buffer = NULL;
@@ -684,6 +746,8 @@ impl_Bonobo_ActivationContext_query (PortableServer_Servant servant,
 
 	actx->refs--;
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -734,9 +798,12 @@ ac_context_to_string_array (CORBA_Context context, char **sort_criteria,
 
         context_username = ac_CORBA_Context_get_value (context, "username", ev);
         context_hostname = ac_CORBA_Context_get_value (context, "hostname", ev);
-	if (ev->_major != CORBA_NO_EXCEPTION)
+	if (ev->_major != CORBA_NO_EXCEPTION) {
+                g_free (context_username);
+                g_free (context_hostname);
                 return;
-        
+        }
+
         sort_criteria[0] = g_strconcat ("username == \'", context_username, "\'", NULL);
         sort_criteria[1] = g_strconcat ("hostname == \'", context_hostname, "\'", NULL);
         sort_criteria[2] = NULL;
@@ -755,13 +822,17 @@ impl_Bonobo_ActivationContext_activateFromAidFull (PortableServer_Servant  serva
                                                    CORBA_Context           ctx,
                                                    CORBA_Environment      *ev)
 {
-        ActivationContext *actx = ACTIVATION_CONTEXT (servant);
-	Bonobo_ActivationResult *retval;
+        ActivationContext *actx;
+	Bonobo_ActivationResult *retval = NULL;
         char *requirements;
         char *sort_criteria[3];
         Bonobo_StringList selection_order;
 	Bonobo_ActivationEnvironment environment;
 
+        server_lock ();
+
+        actx = ACTIVATION_CONTEXT (servant);
+
 	if (strncmp ("OAFAID:", aid, 7) != 0) {
 		Bonobo_Activation_ParseFailed *ex;
 
@@ -771,26 +842,26 @@ impl_Bonobo_ActivationContext_activateFromAidFull (PortableServer_Servant  serva
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
 				     ex_Bonobo_Activation_ParseFailed,
 				     ex);
-		return NULL;
+                goto act_out;
 	}
 
 	ac_update_lists (actx, ev);
         if (ev->_major != CORBA_NO_EXCEPTION)
-                return NULL;
+                goto act_out;
 
 	actx->refs++;
 
         requirements = ac_aid_to_query_string (aid);
         if (requirements == NULL) {
                 actx->refs--;
-                return NULL;
+                goto act_out;
         }
 
         ac_context_to_string_array (ctx, sort_criteria, ev);
         if (ev->_major != CORBA_NO_EXCEPTION) {
                 actx->refs--;
                 g_free (requirements);
-                return NULL;
+                goto act_out;
         }
 
         selection_order._length = 2;
@@ -809,6 +880,9 @@ impl_Bonobo_ActivationContext_activateFromAidFull (PortableServer_Servant  serva
 
         actx->refs--;
 
+ act_out:
+        server_unlock ();
+
         return retval;
 }
 
@@ -839,7 +913,8 @@ activation_context_setup (PortableServer_POA     poa,
                           Bonobo_ObjectDirectory dir,
                           CORBA_Environment     *ev)
 {
-        main_ac = g_object_new (activation_context_get_type (), NULL);
+        main_ac = g_object_new (activation_context_get_type (),
+                                "poa", poa, NULL);
 
         impl_Bonobo_ActivationContext_addDirectory
                 (BONOBO_OBJECT (main_ac), dir, ev);
