@@ -55,6 +55,22 @@ typedef struct {
 	} servers [1]; /* flexible array */
 } ActiveServerList;
 
+typedef struct {
+          /* client's environment */
+        Bonobo_ActivationEnvironment *env;
+          /* "runtime" servers registered by this client */
+        CORBA_sequence_CORBA_string *runtime_iids;
+} ClientContext;
+
+static void
+client_context_free (ClientContext *self)
+{
+        CORBA_free (self->env);
+        if (self->runtime_iids)
+                CORBA_free (self->runtime_iids);
+        g_free (self);
+}
+
 static ObjectDirectory *main_dir = NULL;
 
 #ifdef BONOBO_ACTIVATION_DEBUG
@@ -254,8 +270,12 @@ impl_Bonobo_ObjectDirectory__get_servers (
         Bonobo_CacheTime                 only_if_newer,
         CORBA_Environment               *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
-	Bonobo_ServerInfoListCache      *retval;
+	ObjectDirectory *od;
+	Bonobo_ServerInfoListCache *retval;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
         update_registry (od, FALSE);
 
@@ -268,6 +288,8 @@ impl_Bonobo_ObjectDirectory__get_servers (
 					    CORBA_FALSE);
 	}
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -288,8 +310,12 @@ impl_Bonobo_ObjectDirectory_get_active_servers (
         Bonobo_CacheTime                 only_if_newer,
         CORBA_Environment               *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
-	Bonobo_ServerStateCache         *retval;
+	ObjectDirectory *od;
+	Bonobo_ServerStateCache *retval;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
 	retval = Bonobo_ServerStateCache__alloc ();
 
@@ -310,6 +336,8 @@ impl_Bonobo_ObjectDirectory_get_active_servers (
 					    CORBA_TRUE);
 	}
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -389,8 +417,8 @@ od_merge_client_environment (ObjectDirectory                    *od,
                 goto exit;
 
         /* do the actual merging */
-        client_env = (const Bonobo_ActivationEnvironment *)
-                g_hash_table_lookup (od->client_envs, client);
+        client_env = ((ClientContext *) g_hash_table_lookup
+                      (od->client_contexts, client))->env;
 
         if (!client_env)
                 goto exit;
@@ -438,14 +466,16 @@ impl_Bonobo_ObjectDirectory_activate (
 	CORBA_Context                       ctx,
 	CORBA_Environment                  *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
-	CORBA_Object                     retval;
-	Bonobo_ServerInfo               *si;
-	ODActivationInfo                 ai;
-#ifdef BONOBO_ACTIVATION_DEBUG
-	static int                       depth = 0;
-#endif
-        Bonobo_ActivationEnvironment merged_environment;
+	ObjectDirectory   *od;
+	CORBA_Object       retval = NULL;
+	Bonobo_ServerInfo *si;
+	ODActivationInfo   ai;
+
+        Bonobo_ActivationEnvironment merged_environment = { 0, };
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
         od_merge_client_environment (od, (Bonobo_ServerInfo *)
                                      g_hash_table_lookup (od->by_iid, iid),
@@ -458,24 +488,16 @@ impl_Bonobo_ObjectDirectory_activate (
         if (!(flags & Bonobo_ACTIVATION_FLAG_PRIVATE)) {
                 retval = od_get_active_server (od, iid, &merged_environment);
 
-                if (retval != CORBA_OBJECT_NIL) {
-                        g_free (merged_environment._buffer);
-                        return retval;
-                }
+                if (retval != CORBA_OBJECT_NIL)
+                        goto act_out;
         }
 
-	if (flags & Bonobo_ACTIVATION_FLAG_EXISTING_ONLY) {
-		return CORBA_OBJECT_NIL;
-        }
+	if (flags & Bonobo_ACTIVATION_FLAG_EXISTING_ONLY)
+                goto act_out;
 
 #ifdef BONOBO_ACTIVATION_DEBUG
-        {
-                int i;
-                depth++;
-                for (i = 0; i < depth; i++)
-                        fputc (' ', stderr);
-                fprintf (stderr, "Activate '%s'\n", iid);
-        }
+        fprintf (stderr, "thread %p start activate '%s'\n",
+                 g_thread_self(), iid);
 #endif
 
 	ai.ac = ac;
@@ -487,6 +509,7 @@ impl_Bonobo_ObjectDirectory_activate (
 	if (si) {
 		retval = od_server_activate (
 				si, &ai, BONOBO_OBJREF (od), &merged_environment, client, ev);
+                /* NB. si can now be invalid - due to re-enterancy */
 
                 /* If we failed to activate - it may be because our
                  * request re-entered _during_ the activation
@@ -512,16 +535,17 @@ impl_Bonobo_ObjectDirectory_activate (
         }
 
 #ifdef BONOBO_ACTIVATION_DEBUG
-        {
-                int i;
-                for (i = 0; i < depth; i++)
-                        fputc (' ', stderr);
-                fprintf (stderr, "Activated '%s' = %p\n", iid, retval);
-                depth--;
-        }
+        fprintf (stderr, "thread %p end activate '%s' = %p ['%s']\n",
+                 g_thread_self(), iid, retval,
+                 bonobo_exception_get_text (ev)
+                 );
 #endif
+
+ act_out:
         g_free (merged_environment._buffer);
 
+        server_unlock ();
+
 	return retval;
 }
 
@@ -647,15 +671,28 @@ prune_dead_servers (gpointer key,
         return !servers->n_servers;
 }
 
-static void
-active_server_cnx_broken (ORBitConnection *cnx,
-                          gpointer         dummy)
-
+static gboolean
+as_rescan (gpointer data)
 {
-        ObjectDirectory *od = main_dir;
+        ObjectDirectory *od;
+        static gboolean in_rescan = FALSE;
+        static guint idle_id = 0;
 
-        if (!od) /* shutting down */
-                return;
+        server_lock ();
+
+        if (!(od = main_dir)) { /* shutting down */
+                server_unlock ();
+                return FALSE;
+        }
+
+        /* We tend to get a lot of 'broken' callbacks at once */
+        if (in_rescan) {
+                if (!idle_id)
+                        idle_id = g_idle_add (as_rescan, NULL);
+                server_unlock ();
+                return FALSE;
+        }
+        in_rescan = TRUE;
 
         g_hash_table_foreach_remove (od->active_server_lists,
                                      prune_dead_servers, od);
@@ -665,6 +702,19 @@ active_server_cnx_broken (ORBitConnection *cnx,
 #endif
 
         check_quit ();
+
+        in_rescan = FALSE;
+        server_unlock ();
+
+        return FALSE;
+}
+
+static void
+active_server_cnx_broken (ORBitConnection *cnx,
+                          gpointer         dummy)
+
+{
+        as_rescan (NULL);
 }
 
 static void
@@ -682,10 +732,8 @@ add_active_server (ObjectDirectory    *od,
                         g_object_set_data (
                                 G_OBJECT (cnx), "object_count", GUINT_TO_POINTER (1));
                         
-                        g_signal_connect (
-                                cnx, "broken",
-                                G_CALLBACK (active_server_cnx_broken),
-                                NULL);
+                        ORBit_small_listen_for_broken
+                                (object, G_CALLBACK (active_server_cnx_broken), NULL);
                 }
         } else
                 g_assert (!strcmp (iid, NAMING_CONTEXT_IID) ||
@@ -777,13 +825,15 @@ remove_active_server (ObjectDirectory *od,
   /* Parse server description and register it, replacing older
    * definition if necessary.  Returns the regsitered ServerInfo */
 static Bonobo_ServerInfo const *
-od_register_runtime_server_info (ObjectDirectory  *od,
-                                 const char       *iid,
-                                 const CORBA_char *description)
+od_register_runtime_server_info (ObjectDirectory         *od,
+                                 const char              *iid,
+                                 const CORBA_char        *description,
+                                 Bonobo_ActivationClient  client)
 {
         Bonobo_ServerInfo *old_serverinfo, *new_serverinfo;
         GSList *parsed_serverinfo = NULL, *l;
         int     i;
+        ClientContext *context;
 
         update_registry (od, FALSE);
 
@@ -802,7 +852,9 @@ od_register_runtime_server_info (ObjectDirectory  *od,
                  return NULL;
            /* check for more than one entry */
          if (parsed_serverinfo->next) {
+#ifdef BONOBO_ACTIVATION_DEBUG
                  g_warning ("More than one <oaf_server> specified, ignoring all");
+#endif
                  for (l = parsed_serverinfo; l; l = l->next) {
                          Bonobo_ServerInfo__freekids (l->data, NULL);
                          g_free (l->data);
@@ -825,6 +877,21 @@ od_register_runtime_server_info (ObjectDirectory  *od,
                  g_hash_table_insert (od->by_iid,
                                       od->attr_servers->_buffer[i].iid,
                                       od->attr_servers->_buffer + i);
+ 
+           /* Take note that this client registered this iid, so that
+            * when the client disconnects we unregister its
+            * corresponding serverinfo's */
+         context = g_hash_table_lookup (od->client_contexts, client);
+         g_return_val_if_fail (context, new_serverinfo);
+         if (context->runtime_iids)
+                 ORBit_sequence_append (context->runtime_iids, iid);
+         else {
+                 context->runtime_iids = ORBit_sequence_alloc 
+                         (TC_CORBA_sequence_CORBA_string, 1);
+                 ORBit_sequence_index (context->runtime_iids, 0) =
+                         CORBA_string_dup (iid);
+         }
+
          od->time_list_changed = time (NULL);
          activation_clients_cache_notify ();
          return new_serverinfo;
@@ -842,30 +909,38 @@ impl_Bonobo_ObjectDirectory_register_new_full (
         Bonobo_ActivationClient             client,
 	CORBA_Environment                  *ev)
 {
-	ObjectDirectory              *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory              *od;
 	CORBA_Object                  oldobj;
         Bonobo_ActivationEnvironment  merged_environment;
         Bonobo_ServerInfo const      *serverinfo;
+        Bonobo_RegistrationResult     retval = Bonobo_ACTIVATION_REG_SUCCESS;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
 	oldobj = od_get_active_server (od, iid, environment);
         *existing = oldobj;
 
-        serverinfo = od_register_runtime_server_info (od, iid, description);
+        serverinfo = od_register_runtime_server_info (od, iid, description, client);
         od_merge_client_environment (od, serverinfo, environment,
                                      &merged_environment, client);
 
 	oldobj = od_get_active_server (od, iid, &merged_environment);
 	if (oldobj != CORBA_OBJECT_NIL) {
 		if (!CORBA_Object_non_existent (oldobj, ev)) {
-                        g_free (merged_environment._buffer);
-			return Bonobo_ACTIVATION_REG_ALREADY_ACTIVE;
+                        if (CORBA_Object_is_equivalent (oldobj, obj, ev))
+                                retval = Bonobo_ACTIVATION_REG_SUCCESS;
+                        else
+                                retval = Bonobo_ACTIVATION_REG_ALREADY_ACTIVE;
+                        goto reg_out;
                 }
 	}
 
         if (!serverinfo) {
                 if (!(flags&Bonobo_REGISTRATION_FLAG_NO_SERVERINFO)) {
-                        g_free (merged_environment._buffer);
-                        return Bonobo_ACTIVATION_REG_NOT_LISTED;
+                        retval = Bonobo_ACTIVATION_REG_NOT_LISTED;
+                        goto reg_out;
                 }
         }
 
@@ -874,14 +949,17 @@ impl_Bonobo_ObjectDirectory_register_new_full (
 #endif
 
         add_active_server (od, iid, &merged_environment, obj);
-        g_free (merged_environment._buffer);
 	
 	bonobo_event_source_notify_listeners
                 (od->event_source,
                  "Bonobo/ObjectDirectory:activation:register",
                  NULL, ev);
 
-	return Bonobo_ACTIVATION_REG_SUCCESS;
+ reg_out:
+        g_free (merged_environment._buffer);
+        server_unlock ();
+
+	return retval;
 }
 
 static Bonobo_RegistrationResult
@@ -907,7 +985,11 @@ impl_Bonobo_ObjectDirectory_unregister (
 	const CORBA_Object      obj,
 	CORBA_Environment      *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory *od;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
         if (!remove_active_server (od, iid, obj))
                 CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
@@ -918,6 +1000,8 @@ impl_Bonobo_ObjectDirectory_unregister (
                         (od->event_source,
                          "Bonobo/ObjectDirectory:activation:unregister",
                          NULL, ev);
+
+        server_unlock ();
 }
 
 static Bonobo_DynamicPathLoadResult 
@@ -926,20 +1010,27 @@ impl_Bonobo_ObjectDirectory_add_path(
 	const CORBA_char *		add_path,
 	CORBA_Environment               *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory *od;
 	int i, j, dir_num, max;
 	char **add_directoies, **ret;
 	GSList *l, *tmp = NULL;
 	gboolean different;
+        Bonobo_DynamicPathLoadResult retval = Bonobo_DYNAMIC_LOAD_SUCCESS;
+
+        server_lock ();
+
+        od = OBJECT_DIRECTORY (servant);
 
 	if (!od->registry_source_directories) {
 		od->registry_source_directories = split_path_unique (add_path);
-		return Bonobo_DYNAMIC_LOAD_SUCCESS;
+                goto add_path_out;
 	} else
 		add_directoies = split_path_unique (add_path);
 
-	if (!add_directoies)
-		return Bonobo_DYNAMIC_LOAD_ERROR;
+	if (!add_directoies) {
+		retval = Bonobo_DYNAMIC_LOAD_ERROR;
+                goto add_path_out;
+        }
 
 	for (max = i = 0; od->registry_source_directories[i]; i++) {
 		tmp = g_slist_append(tmp,g_strdup(od->registry_source_directories[i]));
@@ -965,7 +1056,8 @@ impl_Bonobo_ObjectDirectory_add_path(
 	if (max == dir_num) {
 		g_strfreev(add_directoies);
 		g_slist_free(tmp);
-		return Bonobo_DYNAMIC_LOAD_ALREADY_LISTED;
+		retval = Bonobo_DYNAMIC_LOAD_ALREADY_LISTED;
+                goto add_path_out;
 	}
 
 	ret = g_new(char *, max + 1);
@@ -980,7 +1072,11 @@ impl_Bonobo_ObjectDirectory_add_path(
 
 	od->registry_source_directories = ret;
 	update_registry(od, TRUE);	
-	return Bonobo_DYNAMIC_LOAD_SUCCESS;
+
+ add_path_out:
+        server_unlock ();
+
+	return retval;
 }
 
 static Bonobo_DynamicPathLoadResult 
@@ -989,15 +1085,22 @@ impl_Bonobo_ObjectDirectory_remove_path(
         const CORBA_char *              remove_path,
         CORBA_Environment               *ev)
 {
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory *od;
 	char **remove_directoies, **ret;
 	int i, j, max;
 	GSList *l, *tmp = NULL;
 	gboolean different;
+        Bonobo_DynamicPathLoadResult retval = Bonobo_DYNAMIC_LOAD_SUCCESS;
+
+        server_lock();
+
+        od = OBJECT_DIRECTORY (servant);
 
 	remove_directoies = split_path_unique (remove_path);
-	if (!remove_directoies)
-		return Bonobo_DYNAMIC_LOAD_ERROR;
+	if (!remove_directoies) {
+                retval = Bonobo_DYNAMIC_LOAD_ERROR;
+		goto rm_path_out;
+        }
 
 	for (max = i = 0; od->registry_source_directories[i]; i++) {
 		different = TRUE;
@@ -1017,7 +1120,8 @@ impl_Bonobo_ObjectDirectory_remove_path(
 	if (max == i) {
 		g_slist_free(tmp);
 		g_strfreev(remove_directoies);
-		return Bonobo_DYNAMIC_LOAD_NOT_LISTED;
+                retval = Bonobo_DYNAMIC_LOAD_NOT_LISTED;
+		goto rm_path_out;
 	}	
 	ret = g_new(char *, max + 1);
 	for (l = tmp, i = 0; l; l = l->next)
@@ -1030,7 +1134,9 @@ impl_Bonobo_ObjectDirectory_remove_path(
 
 	od->registry_source_directories = ret;
 	update_registry(od, TRUE);
-	return Bonobo_DYNAMIC_LOAD_SUCCESS;
+ rm_path_out:
+        server_unlock();
+	return retval;
 }
 
 
@@ -1038,10 +1144,67 @@ static void
 client_cnx_broken (ORBitConnection *cnx,
                    const Bonobo_ActivationClient  client)
 {
-        ObjectDirectory *od = main_dir;
-        if (!od) /* shutting down */
+        ObjectDirectory *od;
+        ClientContext *context;
+        int i;
+
+        server_lock();
+
+        if (!(od = main_dir)) { /* shutting down */
+                server_unlock();
                 return;
-        g_hash_table_remove (od->client_envs, client);
+        }
+
+        /* unregister runtime server definitions */
+        context = g_hash_table_lookup (od->client_contexts, client);
+        if (context->runtime_iids) {
+                for (i = 0; i < context->runtime_iids->_length; ++i)
+                {
+                        CORBA_char *iid = ORBit_sequence_index
+                                (context->runtime_iids, i);
+                        int j;
+
+#ifdef BONOBO_ACTIVATION_DEBUG
+                        fprintf (stderr, "Removing runtime definition '%s' from hash table\n", iid);
+#endif
+                        g_hash_table_remove (od->by_iid, iid);
+                        for (j = 0; j < od->attr_runtime_servers->len; ++j) {
+                                Bonobo_ServerInfo *server =  g_ptr_array_index
+                                        (od->attr_runtime_servers, j);
+                                if (strcmp (server->iid, iid) == 0) {
+#ifdef BONOBO_ACTIVATION_DEBUG
+                                        fprintf (stderr, "Removing from od->attr_runtime_servers[%i]\n", j);
+#endif
+                                        g_ptr_array_remove_index
+                                                (od->attr_runtime_servers, j);
+                                        Bonobo_ServerInfo__freekids (server, NULL);
+                                        g_free (server);
+                                        break;
+                                }
+                        }
+                        for (j = 0; j < od->attr_servers->_length; ++j) {
+                                Bonobo_ServerInfo *server =
+                                        &ORBit_sequence_index (od->attr_servers, j);
+                                if (strcmp (server->iid, iid) == 0) {
+#ifdef BONOBO_ACTIVATION_DEBUG
+                                        fprintf (stderr, "Removing from od->attr_servers[%i]\n", j);
+#endif
+                                        ORBit_sequence_remove (od->attr_servers, j);
+                                        break;
+                                }
+                        }
+
+#ifdef BONOBO_ACTIVATION_DEBUG
+                        fprintf (stderr, "Runtime definition '%s' cleaned\n", iid);
+#endif
+                        
+                }
+        }
+        g_hash_table_remove (od->client_contexts, client);
+        od->time_list_changed = time (NULL);
+        activation_clients_cache_notify ();
+
+        server_unlock();
 }
 
 static void
@@ -1052,8 +1215,13 @@ impl_Bonobo_ObjectDirectory_addClientEnv (
         CORBA_Environment             *ev)
 {
         Bonobo_ActivationEnvironment *env;
-	ObjectDirectory *od = OBJECT_DIRECTORY (servant);
+	ObjectDirectory *od;
+        ClientContext *context;
         int i;
+
+        server_lock();
+
+        od = OBJECT_DIRECTORY (servant);
         
         env = Bonobo_ActivationEnvironment__alloc ();
         env->_length  = env->_maximum = client_env->_length;
@@ -1080,10 +1248,16 @@ impl_Bonobo_ObjectDirectory_addClientEnv (
                 env->_buffer[i].flags = 0;
         }
 
-        g_hash_table_insert (od->client_envs, client, env);
+        context = g_new (ClientContext, 1);
+        context->env = env;
+        context->runtime_iids = NULL;
+
+        g_hash_table_insert (od->client_contexts, client, context);
 
         ORBit_small_listen_for_broken (client, G_CALLBACK (client_cnx_broken),
                                        (gpointer) client);
+
+        server_unlock();
 }
 
 
@@ -1112,7 +1286,8 @@ bonobo_object_directory_init (PortableServer_POA poa,
 {
         g_assert (main_dir == NULL);
 
-        main_dir = g_object_new (OBJECT_TYPE_DIRECTORY, NULL);
+        main_dir = g_object_new (OBJECT_TYPE_DIRECTORY,
+                                 "poa", poa, NULL);
 
         main_dir->registry_source_directories = split_path_unique (registry_path);
         update_registry (main_dir, FALSE);
@@ -1134,6 +1309,8 @@ bonobo_object_directory_re_check_fn (const Bonobo_ActivationEnvironment *environ
 {
         CORBA_Object retval;
 
+        server_lock ();
+
         retval = od_get_active_server (
                 main_dir, (Bonobo_ImplementationID) act_iid, environment);
 
@@ -1160,6 +1337,8 @@ bonobo_object_directory_re_check_fn (const Bonobo_ActivationEnvironment *environ
                 retval = CORBA_OBJECT_NIL;
         }
 
+        server_unlock ();
+
         return retval;
 }
 
@@ -1183,9 +1362,9 @@ object_directory_finalize (GObject *object)
 
         g_strfreev (od->registry_source_directories);
 
-        if (od->client_envs) {
-                g_hash_table_destroy (od->client_envs);
-                od->client_envs = NULL;
+        if (od->client_contexts) {
+                g_hash_table_destroy (od->client_contexts);
+                od->client_contexts = NULL;
         }
 
         parent_class->finalize (object);
@@ -1228,9 +1407,9 @@ object_directory_init (ObjectDirectory *od)
         od->attr_runtime_servers = g_ptr_array_new ();
 	
 	od->event_source = bonobo_event_source_new ();
-        od->client_envs = g_hash_table_new_full
+        od->client_contexts = g_hash_table_new_full
                 (NULL, NULL, NULL,
-                 (GDestroyNotify) CORBA_free);
+                 (GDestroyNotify) client_context_free);
 }
 
 BONOBO_TYPE_FUNC_FULL (ObjectDirectory,
